#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ptrace.h>
#include <poll.h>
#include <errno.h>

#ifndef FALSE
#define FALSE 0
#define TRUE (!FALSE)
#endif

enum types {
   type_byte,
   type_fdset,
   type_pollfd,
};

main(int argc,char **argv)
{
    const char *arg0=argv[0];
    int mem_fd,pid,j,first;
    enum types type=type_byte;
    unsigned long offset,count,i,n;
    char mem_file_name[64];
    unsigned char *buf;
    fd_set *fds;
    struct pollfd *pfd;
    if (argc>3 && !strcmp(argv[1],"-t"))
    {
	if (!strcmp(argv[2],"byte"))
	    type=type_byte;
	else if (!strcmp(argv[2],"fdset"))
	    type=type_fdset;
	else if (!strcmp(argv[2],"pollfd"))
	    type=type_pollfd;
	else
	{
	    fprintf(stderr,"Unknown type: %s",argv[2]);
	    exit(1);
	}
	argc-=2;
	argv+=2;
    }
    if (argc!=4)
    {
	fprintf(stderr,"Usage: %s [opts] <pid> <offset> <count>\n",arg0);
	exit(1);
    }
    pid=atoi(argv[1]);
    if (pid<=0)
    {
	fprintf(stderr,"Illegal pid: %d\n",pid);
	exit(1);
    }
    offset=strtoul(argv[2],NULL,0);
    if (!offset)
    {
	fprintf(stderr,"Illegal offset: %s\n",argv[2]);
	exit(1);
    }
    count=strtoul(argv[3],NULL,0);
    if (!count)
    {
	fprintf(stderr,"Illegal count: %s\n",argv[3]);
	exit(1);
    }
    switch(type)
    {
	case type_byte:
	    n=1;
	    break;
	case type_fdset:
	    n=sizeof(fd_set);
	    break;
	case type_pollfd:
	    n=sizeof(struct pollfd);
	    break;
    }
    n*=count;
    buf=malloc(n);
    if (!buf)
    {
	fprintf(stderr,"Not enough memory to read %d bytes\n",n);
	exit(1);
    }
    sprintf(mem_file_name,"/proc/%d/mem",pid);
    mem_fd=open(mem_file_name,O_RDONLY);
    if (ptrace(PTRACE_ATTACH,pid,NULL,NULL)<0)
    {
	fprintf(stderr,"ptrace attach: %s\n",strerror(errno));
	exit(1);
    }
    waitpid(pid,NULL,0);
    lseek(mem_fd,offset,SEEK_SET);
    read(mem_fd,buf,n);
    if (ptrace(PTRACE_DETACH,pid,NULL,NULL)<0)
    {
	fprintf(stderr,"ptrace detach: %s\n",strerror(errno));
	exit(1);
    }
    close(mem_fd);
    first=TRUE;
    for(i=0;i<count;i++)
    {
	switch(type)
	{
	    case type_byte:
		printf("0x%X\n",buf[i]);
		break;
	    case type_fdset:
		fds=((fd_set *)buf)+i;
		printf("[");
		for(j=0;j<FD_SETSIZE;j++)
		    if (FD_ISSET(j,fds))
		    {
			if (!first)
			    putchar(' ');
			printf("%d",j);
			first=FALSE;
		    }
		printf("]\n");
		break;
	    case type_pollfd:
		pfd=((struct pollfd *)buf)+i;
		printf("{fd=%d, events=%d, revents=%d}\n",pfd->fd,pfd->events,
		  pfd->revents);
		break;
	}
    }
    free(buf);
}
